import numpy as np

from c2s2.base.cluster import ClusteringAlgorithm
from sklearn import cluster


class DBScan(ClusteringAlgorithm):
    def cluster(self, similarity_matrix, k_clusters):
        """Cluster the similarity matrix
        :param similarity_matrix: the similarity matrix of the patients to cluster
        :param k_clusters: the number of clusters to form
        """
        DBSCANClusterer = cluster.DBSCAN(metric='precomputed')
        #DBSCAN implementation needs distance matrix, we have similarity, so lets flip distribution
        distance_matrix = np.max(similarity_matrix) - similarity_matrix
        DBSCANClusterer.fit(distance_matrix)
        return DBSCANClusterer.labels_


class SpectralClustering(ClusteringAlgorithm):
    def cluster(self, similarity_matrix, k_clusters):
        """Cluster the similarity matrix
        :param similarity_matrix: the similarity matrix of the patients to cluster
        :param k_clusters: the number of clusters to form
        """
        SpectralClusterer = cluster.SpectralClustering(n_clusters=k_clusters, affinity='precomputed')
        SpectralClusterer.fit(similarity_matrix)
        return SpectralClusterer.labels_




import random
import networkx as nx

from c2s2.base.perturb import Perturbation


class Imprecision(Perturbation):
    def __init__(self, number_of_terms_to_replace, hpo_graph):
        """Constructor
        :param number_of_terms_to_replace: number of HPO terms to replace by (grand)parent terms
        :param hpo_graph: the HPO graph
        """
        self._hpo_graph = hpo_graph
        self._number_of_terms_to_replace = number_of_terms_to_replace

    def perturb(self, patient_list):
        """Perturb the input data
        :param patient_list: could be a dataframe with a column of HPO ids as lists, a spark dataframe, phenopackets, etc
        """
        # execution time of this class for 131072 individuals: 7.0 seconds (excluding construction of HPO graph)
        new_patient_list = patient_list[:]
        for i in range(len(new_patient_list)):
            terms_to_replace = [new_patient_list[i][y] for y in sorted(random.sample(range(len(new_patient_list[i])),
                                                                                     self._number_of_terms_to_replace))]
            for term in terms_to_replace:
                all_ancestors = nx.ancestors(self._hpo_graph, term)
                new_term = random.sample(list(all_ancestors), 1)[0]
                new_patient_list[i] = [new_term if a == term else a for a in new_patient_list[i]]
        return new_patient_list




import json
from urllib.request import urlopen
import networkx as nx


def get_hpo_graph_in_networkx(hpo_url):
    """Obtain the HPO graph in networkx format
    :param hpo_url: URL to the HPO JSON file
    """
    with urlopen(hpo_url, timeout=30) as uh:
        data = uh.read()
        hpo_json = json.loads(data.decode("utf-8"))

    edge_list_in_json = hpo_json['graphs'][0]['edges']

    edge_list = []
    for edge in edge_list_in_json:
        # fix formatting to correspond to our wanted format
        # this does include obsolete terms, but does not add synonyms
        node_1 = edge['obj'].replace('http://purl.obolibrary.org/obo/', '').replace('_', ':')
        node_2 = edge['sub'].replace('http://purl.obolibrary.org/obo/', '').replace('_', ':')
        edge_list.append((node_1, node_2))

    return nx.from_edgelist(edge_list, create_using=nx.DiGraph)



import unittest

import numpy as np

from .cluster import DBScan, SpectralClustering


class ClusteringTester(unittest.TestCase):

    def setUp(self):
        random_mat = np.random.uniform(0, 1, (100, 100))
        self._random_symmetric_mat = np.tril(random_mat) + np.tril(random_mat, -1).T
        assert (np.mean(self._random_symmetric_mat == self._random_symmetric_mat.T) == 1)
        self._k_clusters = np.random.randint(2, 10)

    def test_dbscan(self):
        cluster_labels = DBScan().cluster(self._random_symmetric_mat, self._k_clusters)
        assert(len(cluster_labels) == 100)
        assert(max(cluster_labels) < self._k_clusters)

    def test_spectralclustering(self):
        cluster_labels = SpectralClustering().cluster(self._random_symmetric_mat, self._k_clusters)
        assert(len(cluster_labels) == 100)
        assert(max(cluster_labels) < self._k_clusters)
